#define QT_THREAD_SUPPORT
#include <qthread.h>

#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>

#include "rserrno.h"
#include "mcua.h"

#include "exchange.h"
#include "tools.h"
#include "messages.h"
#include "names.h"

extern QMutex	dataMutex;
extern bool		withDevice;

namespace exch {

// lobushka vybrosov:
//	bool vybros = false;

	static unsigned char respbuf[CACHE_MAXSIZE];
	static unsigned char *msgdata;

	static int id_mcua = 0;

	FILE*		srv_file;
	static int	n_read_data;
	static int	block_number;

	int	_maxBlockSize;	// yesli prishel udlinennyi paket, on
				// ne dolzhen byt' dlinneye e
	int	_blockSize0;	// data + HEAD of block

	static int	n_counters = 1;
	Counter*	counters[MAX_N_COUNTERS];
	bool		timeout_flag = false;
	bool		data_cycle_condition = true;
}

void exch::wait( int msec ) {
	if( msec > S2MS ) {
		QString str_add;
		QString str = "Wait " +
			str_add.setNum( msec/S2MS ) + " sec...";
		showMessage( str, msec );
        }
	ulong mksec = ulong(msec*MS2MKS);
	usleep( mksec );
}

void exch::terminate_program( void ) {

	jdebug( "Terminate programm" );

	if( id_mcua > 0 ) close( id_mcua );
//	if( id_data > 0 ) close( id_data );
//	if( id_stat > 0 ) close( id_stat );
//	if( id_proc > 0 ) close( id_proc );

	free( msgdata );
}

int size_data_fifo = 0;
int size_stat_fifo = 0;

void exch::initExchange( void ) {
	char vers[13];

	jdebug( "start of initThreads:" );

	for( int i=0; i<MAX_N_COUNTERS; i++ ) counters[i]=NULL;
	srv_file = NULL;
	n_read_data = 0;
	block_number = 512;

	if( withDevice ) {
		if( (id_mcua = open( "/dev/mcua0", O_RDWR | O_NONBLOCK )) < 0 ) { // 
			switch( errno ) {
				case EHOSTDOWN	: printf("%s\n",NO_DEVICE); break;
				case ESHUTDOWN	: printf("MASS is off or disconnected from LPT converter\n" ); break;
				case ENODEV		: printf("No RS485 driver detected\n"); break;
				default			: perror("?Open mcua"); exit(3);
			}
			printf( "Start without device? (Y/N): " );
			char yesno;
			scanf( "%c",&yesno );
			if( (yesno != 'Y') && (yesno != 'y') ) exit(3);
			else {
				withDevice = false;
				return;
			}
		}
		int verscode = ioctl( id_mcua, RS_GET_VERS, vers );
		if( verscode < 0 ) {	// old driver, sorry...
			printf( "Your rs485 driver is very old... Sorry...\n" );
			exit(3);
		}
		printf( "                   rs485 driver ver.%s", vers);
		bool rtl = RS485_RTL_CODE(verscode);
		if( rtl ) printf( "-rtl\n" );
		else printf("\n");
		size_data_fifo = ioctl( id_mcua, RS_GET_DATAFIFO_SIZE );
		if( size_data_fifo <= 0 ) {
			printf( "DATA_FIFO_SIZE = %d... It's very strange...\n",
						size_data_fifo );
			exit(3);
		} else {
			msgdata = (unsigned char *) malloc(size_data_fifo*sizeof(uchar));
		}
	}
	atexit( terminate_program );
}

void exch::setCounters( Counter** p_counters ) {
	for( int i=0; i<MAX_N_COUNTERS; i++ ) {
		counters[i] = p_counters[i];
	}
}

void exch::setSrvFile( FILE* f ) {
	srv_file = f;
}

int exch::sendCommand( const char* command, bool save=true ) {

	uint ibuf[4];
	uchar *r;

	int key = sscanf( command,"%X %X %X %X",&ibuf[0],
                                        &ibuf[1],&ibuf[2],&ibuf[3] );
	if( key > 0 ) {
		return sendCommand( ibuf[0], ibuf[1],
			&r, key,
			ibuf[2], ibuf[3], save );
	} else return(-1);
}

int exch::sendCommand( uchar address,
			uchar command,
			uchar** r,
			int buf_len=2,
			uchar low=0,
			uchar high=0,
			bool save=true ) {
	*r = respbuf;

	if( buf_len > 4 ) buf_len = 4;
	else if( buf_len < 2 ) buf_len = 2;
	respbuf[0] = address;
	respbuf[1] = command;
	respbuf[2] = low;
	respbuf[3] = high;
		
	if( save && srv_file ) {
		fprintf( srv_file,"-> %X %X",
		(uchar)address,(uchar)command );
		if( buf_len > 2 ) fprintf( srv_file," %X",(uchar)low );
		if( buf_len > 3 ) fprintf( srv_file," %X",(uchar)high );
	}
	int retval = write( id_mcua, respbuf, buf_len ); // 

	if( retval < 0 ) { 
		return retval;
//		printf( "> %s\n",rs_geterror(errno) );
//		if( srv_file ) fprintf( srv_file, " sendCommand: errno=%d\n",errno );
//		exit(3);
	}
	if( retval > CACHE_MAXSIZE ) {	// Signal
//		respbuf[0] = retval;
		respbuf[0] = 0;
		if( save && srv_file )
				fprintf( srv_file,"   <- %s\n",rs_getsignal(retval) );
//		if( retval == SignalDSC ) return(-2);
		switch(retval) {
			case ACY :	return OK;
			case ACW :
			case DSC :	return BUSY;
			default  :	return BUSY;
		}
//		if( retval == SignalDSC ) return BUSY;
//		return(0);
	} else {
		if( save && srv_file ) {
			fprintf( srv_file,"   <-... " );
			for( int i=1; i<retval; i++ ) {
				fprintf( srv_file,"%X ",respbuf[i] );
			}
			fprintf( srv_file,"\n" );
		}
		return DATA;
//		return(1);
	}
}

QString exch::talk( const QString& text ) {
	QString response="";
	int resp = sendCommand( text.data() );
	if( resp == 1 ) {
		for( int i=1; i<respbuf[0]+1; i++ ) {
			QString str;
			str.sprintf( "%X ",respbuf[i] );
			response.append( str );
		}
		return response;
	} else if( (resp == 0) || (resp == -2) ) {
		response.sprintf( "%s",rs_getsignal(respbuf[0]) );
		return response;
	} else {
		response = "Invalid command!";
		return response;
	}
}

int res_read_data = 7;

void exch::*read_data( void ) {		// dataThread

	atexit( quitDataCycle );
// DEBUG !!!
	static QTime t_prev;

	jdebug( "+++++++++ read_data: Start..." );
	n_read_data = 0;

	int count_byte = 0;
	while( data_cycle_condition && withDevice ) {
		while( n_read_data < n_counters*block_number ) {
			if( !data_cycle_condition
					|| !withDevice ) break;
			count_byte = read(id_mcua, msgdata, size_data_fifo);
			if( count_byte == 0 ) {
				usleep( 10000 );
				continue;
			}

// DEBUG!!!
/* */
			QTime dt = QTime::currentTime();
			
			int dmsec = -1*dt.msecsTo(t_prev);
			if( (dmsec>100) && srv_file ) fprintf( srv_file,"big interval %4d msec!\n",
						dmsec );
			t_prev = dt;
/* */

			putData();
/* */

			if( srv_file ) {
				fprintf( srv_file,"%4d  ", dmsec );
				for( int t=0; t<=count_byte; t++ ) {
					fprintf( srv_file,"%02X ",msgdata[t] );
				}
				int blnum = (int)(msgdata[3] + (msgdata[4] << 8));
				fprintf( srv_file," %d %d\n",n_read_data,blnum );
			}

//
			if( msgdata[1] != 0 ) {
				printf( "Stop!!!! signature = %d\n", msgdata[1] );
				sendCommand( "0x1 0x81" );
				exit(1);
			}
/* */
			n_read_data++;
		}
		n_read_data = 0;
		if( dataMutex.locked() ) dataMutex.unlock();
	}
	if( dataMutex.locked() ) dataMutex.unlock();
	jdebug( "Last: dataMutex unlocked" );
	return( &res_read_data );
}

int exch::setBlockSize( int size ) {
	int blockSize = size;
	_blockSize0 = size+HEAD_SIZE;
	_maxBlockSize = _blockSize0 + HEAD_SIZE; //dlia udlinennykh paketov
	if( _maxBlockSize >= size_data_fifo ) {
		_maxBlockSize = size_data_fifo-1;
		_blockSize0 = _maxBlockSize - HEAD_SIZE;
		blockSize = _blockSize0-HEAD_SIZE;
	}
	return blockSize;
}

void exch::setBlockNumber( int number ) {
	block_number = number;
}

int exch::getNblock( void ) {
	if( !timeout_flag ) {
		return n_counters*block_number;
	} else {
		return n_read_data;
	}
}

void exch::setNcounters( int number ) {
	if( number <= MAX_N_COUNTERS ) n_counters = number;
}

int exch::getBlockNumber( void ) {
	return block_number;
}

int exch::getNreadData( void ) {
	return n_read_data;
}

void exch::setTimeout( bool tmout ) {
	timeout_flag = tmout;
}

bool exch::isTimeout( void ) {
	return timeout_flag;
}

void exch::startCounting( void ) {
	n_read_data = 0;
}

void exch::quitDataCycle( void ) {
	jdebug( "quitDataCycle" );
	data_cycle_condition = false;
	if( dataMutex.locked() ) dataMutex.unlock();
	jdebug( "quitDataCycle: dataMutex unlocked" );
}

void exch::timeoutAction( void ) {
	timeout_flag = true;
	if( srv_file ) fprintf( srv_file,"timeout_flag=%d\n",timeout_flag );
	if( dataMutex.locked() ) dataMutex.unlock();
	jdebug( "dataMutex unlocked" );
	if( srv_file ) fprintf( srv_file,"dataMutex unlocked\n" );
/*
// debug!!!
			printf( "Stop!!!! timeout!!!!\n" );
			sendCommand( "0x1 0x81" );
			exit(1);
*/
}

void exch::waitData( void ) {
	if( srv_file ) fprintf( srv_file,"waitData...\n" );
	if( !dataMutex.locked() ) {
		if( srv_file ) fprintf( srv_file,"waitData: dataMutex.lock()-2\n" );
		dataMutex.lock(); // if unlock
		if( srv_file ) fprintf( srv_file,"waitData: dataMutex locked-2\n" );
	}
	dataMutex.lock();
	if( srv_file ) fprintf( srv_file,"waitData: dataMutex locked\n" );
}

/*
void exch::writeToProc( char* to_proc_file, int size ) {
	write( id_proc,to_proc_file,size );
}
*/

void exch::resetBlockCounter(void) {
	ioctl(id_mcua, RS_BLNUMBER_RESET);
}

void exch::putData( void ) {
	Counter *currentCounter = NULL;
	for( int i=0; i<MAX_N_COUNTERS; i++ ) {
		if( counters[i] )
			if( msgdata[2] == (uchar)counters[i]->getModule() )
				currentCounter = counters[i];
	}

	if( !currentCounter ) {
		if( srv_file ) fprintf( srv_file, "Unknown counter with address %X\n",msgdata[2] );
/*
			printf( "Stop!!!!unknown counter!!!\n" );
			sendCommand( "0x1 0x81" );
			exit(1);
*/
		return;
	}

	uchar size = msgdata[0];

	bool ok = true;
	if( size != (_blockSize0-1) ) {
		ok = false;
	}

	int _blnumber = int(msgdata[3] + (msgdata[4] << 8));

	int dnumb = abs( _blnumber - currentCounter->blockCounter.get() );
	if( dnumb != 0 ) {
		if( srv_file ) fprintf( srv_file,"%s block %02X-1 %02X skipped\n",
				currentCounter->getName(),msgdata[3],msgdata[4] );
/*
			printf( "Stop!!!! block skipped!!!\n" );
			sendCommand( "0x1 0x81" );
			exit(1);
*/
		currentCounter->blockCounter.set( _blnumber );
	}

// LOVUSHKA vybrosov:
//	if( currentCounter->getModule()
//			== counters[0]->getModule() ) {
//		vybros = false;
//	}
	if( ok ) {
		currentCounter->addBlockLabel( _blnumber );
		for( uchar t=HEAD_SIZE; t<=size; t=t+2 ) {
			count_t val = msgdata[t]+(msgdata[t+1] << 8);
			currentCounter->putData( val );

//	Yest' vybros???
//			if( currentCounter->getModule()
//				== counters[0]->getModule() ) {
//				if( val >= 1000 ) vybros = true;
//			}
		}
//		if( vybros && srv_file ) {
//			for( int t=0; t<=size; t++ ) {
//				fprintf( srv_file,"%02X ",msgdata[t] );
//			}
//			fprintf( srv_file,"\n" );
//		}
		currentCounter->blockCounter.incr();
	} else {
		jdebug( "!!!!!!!!!!!!!!%s: Skip block !!!!!!!",
				currentCounter->getName() );
		if( srv_file ) fprintf( srv_file,"%s Skip block %02X %02X\n",
				currentCounter->getName(),msgdata[3],msgdata[4] );
	}
}
